/*
 * General functionality supporting file paging.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "file_pager.h"
#include "file_paging.h"

#include <systypes/fcntl.h>



FilePaging::FilePaging(Pages *pages)
: _pages(pages)
{
}

FilePaging::~FilePaging()
{
}



/* Return any registered page mapper for the given 'fileid' or NULL if no such
   mapper is registered. */

PageMapper *FilePaging::get(fileid_t fileid)
{
    FileMapping::iterator entry = _mappers.find(fileid);
    PageMapper *mapper;

    if (entry == _mappers.end())
        mapper = NULL;
    else
        mapper = entry->second;

    return mapper;
}

/* Remove a page mapper and its resources for the given 'fileid'. */

void FilePaging::remove(fileid_t fileid, PageMapper *mapper)
{
    _mappers.erase(fileid);
    delete mapper->accessor();
    delete mapper;
}

/* Register a page 'mapper' for the given 'fileid'. */

void FilePaging::set(fileid_t fileid, PageMapper *mapper)
{
    FileMapping::iterator entry = _mappers.find(fileid);

    if (entry == _mappers.end())
        _mappers[fileid] = mapper;
}



/* Convert opening flags to map-compatible paging flags. */

flags_t FilePaging::get_flags(flags_t flags)
{
    return flags & (O_WRONLY | O_RDWR) ? L4RE_DS_MAP_FLAG_RW : L4RE_DS_MAP_FLAG_RO;
}



/* Obtain a page mapper for the 'fileid' or register a new one in the
   paging object. */

PageMapper *FilePaging::get_mapper(const char *path, fileid_t fileid)
{
    /* Obtain any registered page mapper. */

    PageMapper *mapper = get(fileid);

    if (mapper != NULL)
        return mapper;

    /* Make an accessor and page mapper, registering the mapper. */

    Accessor *accessor = make_accessor(path, fileid);

    if (accessor == NULL)
        return NULL;

    mapper = new PageMapper(accessor, _pages);

    set(fileid, mapper);

    return mapper;
}



/* Return a pager initialised with a page mapper. */

Pager *FilePaging::get_pager(const char *path, fileid_t fileid, flags_t flags)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Initialise the pager with the mapper and a reference to this object for
       closing the mapper and accessor. */

    PageMapper *mapper = get_mapper(path, fileid);

    if (mapper == NULL)
        return NULL;

    return new FilePager(fileid, mapper, flags, this);
}

/* Detach a pager, potentially removing its resources. */

void FilePaging::detach_pager(fileid_t fileid, PageMapper *mapper)
{
    std::lock_guard<std::mutex> guard(_lock);

    if (!mapper->detach())
        remove(fileid, mapper);
}

// vim: tabstop=4 expandtab shiftwidth=4
